// -------------------------------------------------------------------------------
// Image Ehancer                                            Christophe Kohler 2005
//
// Filter that search patterns in pictures and replace them by intermediate colors
// Used for C64 pictures (16 colors)
// Can be used in real time in Vice (done once, but source code has been lost)
//
// This version is a standalone executable.
// It load a TGA picture, process it and save it to TGA file.
// Input picture should be: Any size, RGB (24Bit). TGA format (all pixels one after the others, no header).
// Photoshop can save this.
// Alpha channel is not used.
//
// C64 have two graphic mode:
//  320x200 one color
//  160x200 16 colors (display in 320x200, pixel have 2 width).
// So that's why there is an offset in the cnh_ProcessPixel function (can be 1 pixel or 2 pixel)
// -------------------------------------------------------------------------------

#include <io.h>
#include <sys/stat.h>
#include <direct.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h> // System

#include "ColorEnhancer.h"

void saveTGA(char* name, int sizex, int sizey, unsigned char* buffer);
void FlipColors(int sizex, int sizey, unsigned char* buffer);
void process();
void copyRawToPic();
void copyPicToRaw();
void countcolors();
void OpenInputPicture();
void zoomInputRaw();
void InitColors();
void CheckDoublePixels();

// Class for one pixel (with color R G B A )
class point
{
public:
  point() {}
  ~point() {}
  point(unsigned char _R,unsigned char _G,unsigned char _B,unsigned char _A) {R=_R;G=_G;B=_B;A=_A;}

  unsigned char R,G,B,A;

  point middle(point& p1) { return point( (unsigned char)(0.5f*(R+p1.R)),(unsigned char)(0.5f*(G+p1.G)),(unsigned char)(0.5f*(B+p1.B)),0); };
  int   diff(point& p1) { return abs(R-p1.R)+abs(G-p1.G)+abs(B-p1.B); };
  point & operator = (point& p1) { R=p1.R; G=p1.G; B=p1.B; return (*this);}
  bool operator == (point& p1) { return (R==p1.R) && (G==p1.G) && (B==p1.B); }
  bool operator != (point& p1) { return (R!=p1.R) || (G!=p1.G) || (B!=p1.B); }
};

// Picture data
unsigned char* gSourceRaw; // Temporary files (TGA)
int            gPicWidth;
int            gPicHeight;

unsigned char* gPicSrcIndex; // Image source with index (color index to color table 16 colors)
point          gColortable[256];
int            gNbcolor=0;

point*         gPicSrc;  // Image src, with color as int
point*         gPicDest; // Image dest, with blended colors

unsigned char* gDestRaw;      // Filtered result
unsigned char* gDestRaw_x2;   // Filtered result
unsigned char* gSourceRaw_x2; // Unfiltered result (for quick comparison)

bool           gDoublePixels=false; // Pixel are 2 pixel wide (C64 for example)

char gInputName[512]; // Without extension

// --------------------------------------------------------
// 
// --------------------------------------------------------
int main(int argc, char* argv[])
{
  char fullname[512];

  // Check arguments
  if ( argc!=2 )
  {
    printf("Usage: exe picture.tga\n");
    exit(0);
  }

  FILE* file;

  // Check input file (check if exists)
  file = fopen(argv[1],"rb");
  if (file)
  {
    fclose(file);
  }
  else
  {
    printf("Error: Can not read file %s\n", argv[1]);
    exit(0);
  }

  // Get filename and remove extension
  strcpy(gInputName,argv[1]);
  char* pTmp;
  pTmp = strrchr(gInputName,'.');
  if (pTmp)
    pTmp[0]=0;

  // Build the optimised pattern table (binarized and rotated in all directions)
  // BuildPatternToBlendTable();

  // Open source data (read pixels and put them into SourceRaw array)
  OpenInputPicture(); // Load RGBA picture
  copyRawToPic(); // Copyt picture to our internal format (using point struct)

  countcolors(); // Create indexed image , and count colors

  InitColors();

  CheckDoublePixels(); // Set flag "double pixel" for picture with pixel of 2 width (like C64)

  process();

  copyPicToRaw();

  strcpy(fullname,"result/");
  strcat(fullname,gInputName);
  saveTGA(fullname, gPicWidth, gPicHeight, gSourceRaw);

  strcpy(fullname,"result/");
  strcat(fullname,gInputName);
  strcat(fullname,"_filtered");
  saveTGA(fullname, gPicWidth, gPicHeight, gDestRaw);


  // 
//   char fullname[512];
//   strcpy(fullname,"result/");
//   strcat(fullname,gInputName);
//   strcat(fullname,"_x2_filtered.tga");
//   saveTGA(fullname, 640, 400, DestRaw);
//   
//   zoomInputRaw();
//   strcpy(fullname,"result/");
//   strcat(fullname,gInputName);
//   strcat(fullname,"_x2.tga");
//   saveTGA(fullname, 640, 400, SourceRaw_x2);

  return 0;
}

// --------------------------------------------------------
// Check if input pixutre have pixel with double width 
// C64 pixel are 2 pixel wide. So algorithm need to know where to search next pixel
// --------------------------------------------------------
void CheckDoublePixels()
{
  int x,y;

  gDoublePixels=true;
  // Check all pixels, if two consecutive pixels are not equals, then set gDoublePixels to false and exit

  for (y=0; y<gPicHeight; y++)
  {
    for (x=0; x<gPicWidth/2; x++)
    {
      if ( gPicSrcIndex[y*gPicWidth+(x*2+0)] != gPicSrcIndex[y*gPicWidth+(x*2+1)] )
      {
        gDoublePixels=false;
        return;
      }
    }
  }

}

// --------------------------------------------------------
// Init color for algorythm
// --------------------------------------------------------
void InitColors()
{
  unsigned int Colors[256];

  int i;
  // Build array of color, 1 color = 1 int
  for (i=0; i<gNbcolor; i++)
  {
    Colors[i]=(gColortable[i].R<<16)+(gColortable[i].G<<8)+(gColortable[i].B<<0);
  }
  // Call init of color table for algorithm
  cnh_DefineColorTable((int*)Colors,gNbcolor);
}

// --------------------------------------------------------
// 
// --------------------------------------------------------
int addcolor(point colortoadd)
{
  int i;
  for (i=0; i<gNbcolor; i++)
  {
    if ( colortoadd == gColortable[i] )
      return i; // Color already exist
  }

  // Add color
   if (gNbcolor==256)
   {
     printf("Error: More than 256 colors\n");
     exit(0);
     return 0;
   }

  gColortable[gNbcolor]=colortoadd; // Add color
  gNbcolor++;

  return gNbcolor-1;
}

// --------------------------------------------------------
// 
// --------------------------------------------------------
void countcolors()
{
  // Get color
  int x,y;
  int index;

  // Alloc memory for index buffer
  gPicSrcIndex = (unsigned char*) malloc(gPicWidth*gPicHeight);

  for (y=0; y<gPicHeight; y++)
  {
    for (x=0; x<gPicWidth; x++)
    {
      index=addcolor(gPicSrc[y*gPicWidth+x]);
      gPicSrcIndex[y*gPicWidth+x]=index;
    }
  }
}

// --------------------------------------------------------------------------
// Name: Process
// Process the input picture
// --------------------------------------------------------------------------
void process()
{
  int x,y;
  
  // Create dest picture
  gPicDest=(point*)malloc(gPicWidth*gPicHeight*sizeof(point));

  // Set offset in case of non square pixel (C64 pixels are 2 pixel width)
  int offset;
  if ( gDoublePixels )
    offset = 2; // Next pixel is 2 bytes from current one
  else
    offset = 1; // NExt pixel is 1 byt from current one

  // Copy borders pixels
  if ( !gDoublePixels )
  {
    // Copy one pixel border
    // Top and bottom border
    for (x=0; x<gPicWidth; x++)
    {
      gPicDest[x] = gPicSrc[x];
      gPicDest[(gPicHeight-1)*gPicWidth+x] = gPicSrc[(gPicHeight-1)*gPicWidth+x];
    }
    // Left and right borders
    for (y=1; y<gPicHeight-1; y++)
    {
      gPicDest[y*gPicWidth+0] = gPicSrc[y*gPicWidth+0];
      gPicDest[y*gPicWidth+gPicWidth-1] = gPicSrc[y*gPicWidth+gPicWidth-1];
    }

  }
  else
  {
    // Copy two pixel borders
    // Top and bottom border
    for (x=0; x<gPicWidth; x++)
    {
      gPicDest[x] = gPicSrc[x];
      gPicDest[gPicWidth+x] = gPicSrc[gPicWidth+x];
      gPicDest[(gPicHeight-2)*gPicWidth+x] = gPicSrc[(gPicHeight-2)*gPicWidth+x];
      gPicDest[(gPicHeight-1)*gPicWidth+x] = gPicSrc[(gPicHeight-1)*gPicWidth+x];
    }
    // Left and right borders
    for (y=1; y<gPicHeight-1; y++)
    {
      gPicDest[y*gPicWidth+0] = gPicSrc[y*gPicWidth+0];
      gPicDest[y*gPicWidth+1] = gPicSrc[y*gPicWidth+1];
      gPicDest[y*gPicWidth+gPicWidth-1] = gPicSrc[y*gPicWidth+gPicWidth-1];
      gPicDest[y*gPicWidth+gPicWidth-2] = gPicSrc[y*gPicWidth+gPicWidth-2];
    }
  }
  
  int pitchy;
  pitchy=gPicWidth;

  for (y=offset; y<gPicHeight-offset; y++)
  {
    unsigned char* pSrc;
    pSrc = gPicSrcIndex+y*gPicWidth;

    for (x=offset; x<gPicWidth-offset; x++)
    {
      int destoffset;
      unsigned int color;

      color = cnh_ProcessPixel(pSrc+x,pitchy,offset);

      destoffset = y*gPicWidth+x;
      gPicDest[destoffset].R=(color>>16);
      gPicDest[destoffset].G=(color>>8);
      gPicDest[destoffset].B=(color>>0);
      gPicDest[destoffset].A=0;
    }
  }
}

// --------------------------------------------------------------------------
// Name: copyRawToPic
// Copy input picture (RGBA) to internal format with point structure
// --------------------------------------------------------------------------
void copyRawToPic()
{
  int x,y;
  gPicSrc=(point*)malloc(gPicWidth*gPicHeight*sizeof(point));

  for (y=0; y<gPicHeight; y++)
  {
    for (x=0; x<gPicWidth; x++)
    {
      int offset;
      offset=(y*gPicWidth+x)*4;
      gPicSrc[y*gPicWidth+x].R = gSourceRaw[offset+0];
      gPicSrc[y*gPicWidth+x].G = gSourceRaw[offset+1];
      gPicSrc[y*gPicWidth+x].B = gSourceRaw[offset+2];
      gPicSrc[y*gPicWidth+x].A = gSourceRaw[offset+3];
    }
  }
}

// --------------------------------------------------------------------------
// Name: copyPicToRaw
// --------------------------------------------------------------------------
void copyPicToRaw()
{
  int x,y;

  gDestRaw = (unsigned char*) malloc(gPicWidth*gPicHeight*4);

  for (y=0; y<gPicHeight; y++)
  {
    for (x=0; x<gPicWidth; x++)
    {
      int offsetsrc;
      int offsetdest;
      offsetsrc=(y*gPicWidth+x);
      offsetdest=(y*gPicWidth+x)*4;
      gDestRaw[offsetdest]   = gPicDest[offsetsrc].R;
      gDestRaw[offsetdest+1] = gPicDest[offsetsrc].G;
      gDestRaw[offsetdest+2] = gPicDest[offsetsrc].B;
      gDestRaw[offsetdest+3] = gPicDest[offsetsrc].A;
    }
  }
}

// --------------------------------------------------------------------------
// Name: zoomInputRaw
// Copy inpur raw (320x200) to ouput raw "x2" (640x400)
// --------------------------------------------------------------------------
/*
void zoomInputRaw()
{
  int x,y;

  for (x=0; x<320; x++)
  {
    for (y=0; y<200; y++)
    {
      int offset;
      int offsetsrc;
      offset=(y*2*640+(x*2))*4;
      offsetsrc=(y*320+x)*4;
      SourceRaw_x2[offset]   = SourceRaw[offsetsrc+0];
      SourceRaw_x2[offset+1] = SourceRaw[offsetsrc+1];
      SourceRaw_x2[offset+2] = SourceRaw[offsetsrc+2];
      SourceRaw_x2[offset+3] = SourceRaw[offsetsrc+3];

      SourceRaw_x2[offset  +4] = SourceRaw[offsetsrc+0];;
      SourceRaw_x2[offset+1+4] = SourceRaw[offsetsrc+1];;
      SourceRaw_x2[offset+2+4] = SourceRaw[offsetsrc+2];
      SourceRaw_x2[offset+3+4] = SourceRaw[offsetsrc+3];

      SourceRaw_x2[offset  +(640*4)]   = SourceRaw[offsetsrc+0];
      SourceRaw_x2[offset+1+(640*4)] = SourceRaw[offsetsrc+1];
      SourceRaw_x2[offset+2+(640*4)] = SourceRaw[offsetsrc+2];
      SourceRaw_x2[offset+3+(640*4)] = SourceRaw[offsetsrc+3];

      SourceRaw_x2[offset  +4+(640*4)] = SourceRaw[offsetsrc+0];;
      SourceRaw_x2[offset+1+4+(640*4)] = SourceRaw[offsetsrc+1];;
      SourceRaw_x2[offset+2+4+(640*4)] = SourceRaw[offsetsrc+2];
      SourceRaw_x2[offset+3+4+(640*4)] = SourceRaw[offsetsrc+3];
    }
  }
}
*/

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
void FlipColors(int sizex, int sizey, unsigned char* buffer)
{
  // Flip colors
  // RGBA to BGRA
  // Test
  int i;
  int R, B;
  for (i=0; i<4*sizex*sizey; i+=4)
  {
    R=buffer[i];
    B=buffer[i+2];
   
    buffer[i]=B;
    buffer[i+2]=R;
  }
}

// --------------------------------------------------------------------------
#pragma pack(push, 1)
// tga header
typedef struct
{
  unsigned char id_lenght;          /* size of image id */
  unsigned char colormap_type;      /* 1 is has a colormap */
  unsigned char image_type;         /* compression type */

  signed short	cm_first_entry;       /* colormap origin */
  signed short	cm_length;            /* colormap length */
  unsigned char cm_size;            /* colormap size */

  signed short	x_origin;             /* bottom left x coord origin */
  signed short	y_origin;             /* bottom left y coord origin */

  signed short	width;                /* picture width (in pixels) */
  signed short	height;               /* picture height (in pixels) */

  unsigned char pixel_depth;        /* bits per pixel: 8, 16, 24 or 32 */
  unsigned char image_descriptor;   /* 24 bits = 0x00; 32 bits = 0x80 */

} tga_header_t;
#pragma pack(pop)

// --------------------------------------------------------------------------
// Name : OpenInputPicture
// Read pixels from input. 
// FORMAT TGA
// --------------------------------------------------------------------------
void OpenInputPicture()
{
  FILE*			file;
	char      FullName[256];
  
  strcpy(FullName, gInputName);
  strcat(FullName, ".tga");
  file=fopen(FullName, "rb");

  tga_header_t header;

  // read header 
  fread(&header, sizeof (tga_header_t), 1, file);

  switch (header.image_type)
  { 
  case 2:  /* BGR 16-24-32 bits */
    break;
  default:
    printf ("Error: TGA format not supported\n");
    exit(0);
    break;
  }

  fseek(file, header.id_lenght, SEEK_CUR);

  gPicWidth=header.width;
  gPicHeight=header.height;
  gSourceRaw= (unsigned char*) malloc(gPicWidth*gPicHeight*4);

  // read image data 
  switch (header.image_type)
  {
  case 2:
    // uncompressed 16-24-32 bits 
    switch (header.pixel_depth)
    {
    case 16:
      //ReadTGA16bits (fp, texinfo);
      printf("Error: TGA mode not supported\n");
      exit(0);
      break;

    case 24:
      {
        int i,j;
        // Read pixels and flip vertically the picture (TGA on PC are vertically inverted)
        for (j=gPicHeight-1; j>=0; j--)
        {
          for (i=0; i <gPicWidth; ++i)
          {
            // read and convert BGR to RGB
            unsigned char val;
            fread(&val,1,1,file);
            gSourceRaw[gPicWidth*4*j+i*4+0] = val;
            fread(&val,1,1,file);
            gSourceRaw[gPicWidth*4*j+i*4+1] = val;
            fread(&val,1,1,file);
            gSourceRaw[gPicWidth*4*j+i*4+2] = val;
            gSourceRaw[gPicWidth*4*j+i*4+3] = 255;
          }
        }
      }
      break;

    case 32:
      {
        /*
        int i,j;
        // Read pixels and flip vertically the picture (TGA on PC are vertically inverted)
        for (j=texinfo->height-1; j>=0; j--)
        {
          for (i=0; i <texinfo->width; ++i)
          {
            // read and convert BGRA to RGBA
            GLubyte val;
            fread(&val,1,1,fp);
            texinfo->texels[((j*texinfo->width+i) * texinfo->internalFormat) + 2] = val;
            fread(&val,1,1,fp);
            texinfo->texels[((j*texinfo->width+i) * texinfo->internalFormat) + 1] = val;
            fread(&val,1,1,fp);
            texinfo->texels[((j*texinfo->width+i) * texinfo->internalFormat) + 0] = val;
            fread(&val,1,1,fp);
            texinfo->texels[((j*texinfo->width+i) * texinfo->internalFormat) + 3] = val;
          }
        }
        */
      }
      break;
    }

    break;

  default:
    // image type is not correct 
    printf ("Error: unknown TGA image type %i!\n", header.image_type);
    exit(0);
    break;
  }

  FlipColors(header.width,header.height,gSourceRaw);

  fclose(file);
}

// --------------------------------------------------------------------------
// Name : saveTGA
// Save TGA picture (raw with 18 bytes header)
// FORMAT TGA : BVRA
// --------------------------------------------------------------------------
void saveTGA(char* name, int sizex, int sizey, unsigned char* buffer)
{
	FILE*			fichier;
	unsigned char	mychar;
	unsigned short	myshort;
	char      FullName[256];
  int       compteur=0;
  char      chainecompteur[8];
  fichier=NULL;
  
  strcpy(FullName, name);
  strcat(FullName, ".tga");
  fichier=fopen(FullName, "wb");

  FlipColors(sizex,sizey,buffer);
	
  if (fichier!=NULL)
	{
		// Ecris entete TGA
		mychar=0; // size of ID field that follows 18 byte header (0 usually)
		fwrite(&mychar, 1, 1, fichier);
		mychar=0; // type of colour map 0=none, 1=has palette
		fwrite(&mychar, 1, 1, fichier);
		mychar=2; // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed
		fwrite(&mychar, 1, 1, fichier);
		myshort=0; // first colour map entry in palette
		fwrite(&myshort, 2, 1, fichier);
		myshort=0; // number of colours in palette
		fwrite(&myshort, 2, 1, fichier);
		mychar=0;  // number of bits per palette entry 15,16,24,32
		fwrite(&mychar, 1, 1, fichier);
		myshort=0;  // image x origin
		fwrite(&myshort, 2, 1, fichier);
		myshort=0;  // image y origin
		fwrite(&myshort, 2, 1, fichier);
		myshort=sizex;  // image width in pixels
		fwrite(&myshort, 2, 1, fichier);
		myshort=sizey;  // image height in pixels
		fwrite(&myshort, 2, 1, fichier);
		mychar=32;   // image bits per pixel 8,16,24,32
		fwrite(&mychar, 1, 1, fichier);
		mychar=0x30;   // image descriptor bits (vh flip bits) (00vhaaaa)
		fwrite(&mychar, 1, 1, fichier);

		// Ecrit pixels
		fwrite(buffer, 1, 4*sizex * sizey, fichier);
		fclose(fichier);
	}
	else
	{
		printf("Error, save not possible (%s)\n", name);
	}
}


